"""This file is generated by setuptools-generate.
The information comes from pyproject.toml.
It provide some metainfo for docs/conf.py to build documents and
help2man to build man pages.
"""

# For docs/conf.py
project = "lsp-tree-sitter"
author = """Wu, Zhenyu <wuzhenyu@ustc.edu> """
copyright = "2025"

# For help2man
DESCRIPTION = "a library to create language servers"
EPILOG = "Report bugs to https://github.com/neomutt/lsp-tree-sitter/issues"
# format __version__ by yourself
VERSION = """lsp-tree-sitter {__version__}
Copyright (C) 2025
Written by Wu, Zhenyu <wuzhenyu@ustc.edu> """